import dataclasses
from typing import Optional

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.constraint import command
from pcs.cli.constraint_ticket import parse_args
from pcs.cli.reports.output import (
    deprecation_warning,
    error,
)
from pcs.common import reports
from pcs.common.reports import constraints
from pcs.common.reports.messages import InvalidOptions


def create_with_set(lib, argv, modifiers):
    """
    create ticket constraint with resource set
    object lib exposes library
    list argv see usage for "constraint colocation set"
    dict like object modifiers can contain
        "force" allows resource in clone/master and constraint duplicity

    Options:
      * --force - allow resource inside clone (or master), allow duplicate
        element
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    command.create_with_set(
        lib.constraint_ticket.create_with_set,
        argv,
        modifiers,
    )


def add(lib, argv, modifiers):
    """
    create ticket constraint
    object lib exposes library
    list argv see usage for "constraint colocation add"
    dict like object modifiers can contain
        "force" allows resource in clone/master and constraint duplicity

    Options:
      * --force - allow resource inside clone (or master), allow duplicate
        element
      * -f - CIB file
    """

    def _report_item_preprocessor(
        report_item: reports.ReportItem,
    ) -> Optional[reports.ReportItem]:
        if isinstance(report_item.message, InvalidOptions):
            new_message = dataclasses.replace(
                report_item.message,
                allowed=sorted(set(report_item.message.allowed) - {"rsc-role"}),
            )
            return dataclasses.replace(report_item, message=new_message)
        return report_item

    modifiers.ensure_only_supported("--force", "-f")
    ticket, resource_id, resource_role, options = parse_args.parse_add(argv)
    if "rsc-role" in options:
        raise CmdLineInputError(
            "Resource role must not be specified among options"
            + ", specify it before resource id"
        )

    if resource_role:
        options["rsc-role"] = resource_role

    lib.env.report_processor.set_report_item_preprocessor(
        _report_item_preprocessor
    )

    lib.constraint_ticket.create(
        ticket,
        resource_id,
        options,
        resource_in_clone_alowed=modifiers.get("--force"),
        duplication_alowed=modifiers.get("--force"),
    )


def remove(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if len(argv) != 2:
        raise CmdLineInputError()
    ticket, resource_id = argv
    if not lib.constraint_ticket.remove(ticket, resource_id):
        raise error("no matching ticket constraint found")


def show(lib, argv, modifiers):
    deprecation_warning(
        "This command is deprecated and will be removed. "
        "Please use 'pcs constraint ticket config' instead."
    )
    return config_cmd(lib, argv, modifiers)


def config_cmd(lib, argv, modifiers):
    """
    show all ticket constraints
    object lib exposes library
    list argv see usage for "constraint colocation show"
    dict like object modifiers can contain "full"

    Options:
      * --full - print more details
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f", "--full")
    if argv:
        raise CmdLineInputError()
    print(
        "\n".join(
            command.config_cmd(
                "Ticket Constraints:",
                lib.constraint_ticket.config,
                constraints.ticket_plain,
                modifiers,
            )
        )
    )



import os.path

from pcs import settings
from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.common.str_tools import join_multilines
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner


def get_status_text(runner: CommandRunner, verbose: bool = False) -> str:
    """
    Get quorum device client runtime status in plain text

    verbose -- get more detailed output
    """
    cmd = [
        os.path.join(
            settings.corosync_qdevice_binaries, "corosync-qdevice-tool"
        ),
        "-s",
    ]
    if verbose:
        cmd.append("-v")
    stdout, stderr, retval = runner.run(cmd)
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CorosyncQuorumGetStatusError(
                    join_multilines([stderr, stdout])
                )
            )
        )
    return stdout



from typing import (
    Any,
    Callable,
    Dict,
    List,
    cast,
)

from lxml.etree import (
    SubElement,
    _Element,
)

from pcs.common import reports
from pcs.lib.cib import resource
from pcs.lib.cib.constraint import resource_set
from pcs.lib.cib.tools import (
    find_element_by_tag_and_id,
    find_unique_id,
)
from pcs.lib.errors import LibraryError
from pcs.lib.xml_tools import (
    export_attributes,
    find_parent,
    get_root,
)


def _validate_attrib_names(attrib_names, options):
    invalid_names = [
        name for name in options.keys() if name not in attrib_names
    ]
    if invalid_names:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.InvalidOptions(
                    sorted(invalid_names), sorted(attrib_names), None
                )
            )
        )


def find_valid_resource_id(
    report_processor: reports.ReportProcessor, cib, in_clone_allowed, _id
):
    parent_tags = resource.clone.ALL_TAGS + [resource.bundle.TAG]
    resource_element = find_element_by_tag_and_id(
        sorted(parent_tags + [resource.primitive.TAG, resource.group.TAG]),
        cib,
        _id,
    )

    if resource_element.tag in parent_tags:
        return resource_element.attrib["id"]

    clone = find_parent(resource_element, parent_tags)
    if clone is None:
        return resource_element.attrib["id"]

    report_msg = reports.messages.ResourceForConstraintIsMultiinstance(
        resource_element.attrib["id"],
        "clone" if clone.tag == "master" else clone.tag,
        str(clone.attrib["id"]),
    )
    if in_clone_allowed:
        if report_processor.report(
            reports.ReportItem.warning(report_msg)
        ).has_errors:
            raise LibraryError()
        return resource_element.attrib["id"]

    raise LibraryError(
        reports.ReportItem.error(
            report_msg,
            force_code=reports.codes.FORCE,
        )
    )


def prepare_options(attrib_names, options, create_id_fn, validate_id):
    _validate_attrib_names(attrib_names + ("id",), options)
    options = options.copy()

    if "id" not in options:
        options["id"] = create_id_fn()
    else:
        validate_id(options["id"])
    return options


def export_with_set(element: _Element) -> Dict[str, Any]:
    return {
        "resource_sets": [
            resource_set.export(resource_set_item)
            for resource_set_item in element.findall(".//resource_set")
        ],
        "options": export_attributes(element),
    }


def export_plain(element: _Element) -> Dict[str, Any]:
    return {"options": export_attributes(element)}


def create_id(cib, type_prefix, resource_set_list):
    # Create a semi-random id. We need it to be predictable (for testing), short
    # and somehow different than other ids so that we don't spend much time in
    # find_unique_id.
    # Avoid using actual resource names. It makes the id very long (consider 10
    # or more resources in a set constraint). Also, if a resource is deleted
    # and therefore removed from the constraint, the id no longer matches the
    # constraint.
    resource_ids = []
    for _set in resource_set_list:
        resource_ids.extend(_set["ids"])
    id_part = "".join([_id[0] + _id[-1] for _id in resource_ids][:3])
    return find_unique_id(cib, "{0}_set_{1}".format(type_prefix, id_part))


def have_duplicate_resource_sets(element, other_element):
    def get_id_set_list(element):
        return [
            resource_set.get_resource_id_set_list(resource_set_item)
            for resource_set_item in element.findall(".//resource_set")
        ]

    return get_id_set_list(element) == get_id_set_list(other_element)


def check_is_without_duplication(
    report_processor: reports.ReportProcessor,
    constraint_section: _Element,
    element: _Element,
    are_duplicate: Callable[[_Element, _Element], bool],
    export_element: Callable[[_Element], Dict[str, Any]],
    duplication_allowed: bool = False,
) -> None:
    duplicate_element_list = [
        duplicate_element
        for duplicate_element in cast(
            # The xpath method has a complicated return value, but we know our
            # xpath expression returns only elements.
            List[_Element],
            constraint_section.xpath(
                ".//*[local-name()=$tag_name]", tag_name=element.tag
            ),
        )
        if (
            element is not duplicate_element
            and are_duplicate(element, duplicate_element)
        )
    ]
    if not duplicate_element_list:
        return

    if report_processor.report_list(
        [
            reports.ReportItem.info(
                reports.messages.DuplicateConstraintsList(
                    element.tag,
                    [
                        export_element(duplicate_element)
                        for duplicate_element in duplicate_element_list
                    ],
                )
            ),
            reports.ReportItem(
                severity=reports.item.get_severity(
                    reports.codes.FORCE,
                    duplication_allowed,
                ),
                message=reports.messages.DuplicateConstraintsExist(
                    [
                        str(duplicate.attrib["id"])
                        for duplicate in duplicate_element_list
                    ]
                ),
            ),
        ]
    ).has_errors:
        raise LibraryError()


def create_with_set(constraint_section, tag_name, options, resource_set_list):
    if not resource_set_list:
        raise LibraryError(
            reports.ReportItem.error(reports.messages.EmptyResourceSetList())
        )
    element = SubElement(constraint_section, tag_name)
    element.attrib.update(options)
    if tag_name == "rsc_order":
        all_resource_ids = []
        for resource_set_item in resource_set_list:
            all_resource_ids.extend(resource_set_item["ids"])
        resource_set.is_resource_in_same_group(
            get_root(constraint_section), all_resource_ids
        )
    for resource_set_item in resource_set_list:
        resource_set.create(element, resource_set_item)
    return element



import os.path

from pcs import settings

from pcs_test.tools.command_env.mock_runner import Call as RunnerCall


class ScsiShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def unfence_node(
        self,
        plug,
        devices,
        fence_agent,
        stdout="",
        stderr="",
        return_code=0,
        name="runner.scsi.unfence_node",
    ):
        """
        Create a calls for node scsi/mpath unfencing

        string plug -- a nodename or key for fence_scsi/fence_mpath agents
        list devices -- list of devices to unfence
        string fence_agent -- name of fence agent script
        string stdout -- stdout from fence agent script
        string stderr -- stderr from fence agent script
        int return_code -- return code of the fence agent script
        string name -- the key of this call
        """
        self.__calls.place(
            name,
            RunnerCall(
                [
                    os.path.join(settings.fence_agent_binaries, fence_agent),
                    "--action=on",
                    "--devices",
                    ",".join(sorted(devices)),
                    f"--plug={plug}",
                ],
                stdout=stdout,
                stderr=stderr,
                returncode=return_code,
            ),
        )

    def get_status(
        self,
        node,
        device,
        fence_agent,
        stdout="",
        stderr="",
        return_code=0,
        name="runner.scsi.is_fenced",
    ):
        """
        Create a call for getting scsi status

        string node -- a node from which is unfencing performed
        str device -- a device to check
        string stdout -- stdout from fence_scsi agent script
        string stderr -- stderr from fence_scsi agent script
        int return_code -- return code of the fence_scsi agent script
        string name -- the key of this call
        """
        self.__calls.place(
            name,
            RunnerCall(
                [
                    os.path.join(settings.fence_agent_binaries, fence_agent),
                    "--action=status",
                    f"--devices={device}",
                    f"--plug={node}",
                ],
                stdout=stdout,
                stderr=stderr,
                returncode=return_code,
            ),
        )



from unittest import TestCase

from lxml import etree

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import XmlManipulation

fixture_primitive = """
    <primitive class="ocf" id="A" provider="heartbeat" type="Dummy" />
"""

fixture_constraints = [
    """
    <rsc_location id="cli-ban-A-on-node1" rsc="A" role="Started" node="node1"
        score="-INFINITY"/>
    """,
    """
    <rsc_location id="location-A-node1--INFINITY" rsc="A" node="node1"
        score="-INFINITY"/>
    """,
]

fixture_nodes = [
    """<node id="1" uname="node1"/>""",
    """<node id="2" uname="node2"/>""",
    """<node id="3" uname="node3"/>""",
]


class Move(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    ),
):
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        # pylint: disable=invalid-name
        self.temp_cib = get_tmp_file("tier1_cib_resource_move")
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        xml_manip = XmlManipulation.from_file(self.empty_cib)
        xml_manip.append_to_first_tag_name("resources", fixture_primitive)
        xml_manip.append_to_first_tag_name(
            "constraints",
            *fixture_constraints,
        )
        xml_manip.append_to_first_tag_name(
            "nodes",
            *fixture_nodes,
        )
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def tearDown(self):
        # pylint: disable=invalid-name
        self.temp_cib.close()

    def test_move_to_node_with_existing_ban_constraints(self):
        self.assert_effect(
            "resource move-with-constraint A node1".split(),
            f"<resources>{fixture_primitive}</resources>",
            stderr_full=(
                "Warning: A move constraint has been created and the resource "
                "'A' may or may not move depending on other configuration\n"
            ),
        )
        self.assert_resources_xml_in_cib(
            """
            <constraints>
                <rsc_location id="location-A-node1--INFINITY" rsc="A"
                    node="node1" score="-INFINITY"
                />
                <rsc_location id="cli-prefer-A" rsc="A" role="Started"
                    node="node1" score="INFINITY"
                />
            </constraints>
            """,
            get_cib_part_func=lambda cib: etree.tostring(
                etree.parse(cib).findall(".//constraints")[0]
            ),
        )

    def test_move_stopped_with_existing_constraints(self):
        self.assert_pcs_fail(
            "resource move-with-constraint A".split(),
            (
                "Error: You must specify a node when moving/banning a stopped "
                "resource\n"
            ),
        )

    def test_nonexistent_resource(self):
        self.assert_pcs_fail(
            "resource move-with-constraint NonExistent".split(),
            (
                "Error: bundle/clone/group/resource 'NonExistent' does not "
                "exist\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )

    def test_wait_not_supported_with_file(self):
        self.assert_pcs_fail(
            "resource move-with-constraint A --wait".split(),
            "Error: Cannot use '-f' together with '--wait'\n",
        )

    def test_move_autoclean_not_supported_with_file(self):
        self.assert_pcs_fail(
            "resource move A".split(),
            "Error: Specified option '-f' is not supported in this command\n",
        )



import json
from shlex import split
from unittest import TestCase

from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.resource.list import ListCibResourcesDto

from pcs_test.tools import resources_dto
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class ResourceConfigJson(TestCase):
    def setUp(self):
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-resources.xml"),
        )

    def test_all(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["resource", "config", "--output-format=json"]
        )
        expected = ListCibResourcesDto(
            primitives=[
                resources_dto.PRIMITIVE_R1,
                resources_dto.PRIMITIVE_R7,
                resources_dto.PRIMITIVE_R5,
                resources_dto.STONITH_S1,
                resources_dto.PRIMITIVE_R2,
                resources_dto.PRIMITIVE_R3,
                resources_dto.PRIMITIVE_R4,
                resources_dto.PRIMITIVE_R6,
            ],
            clones=[
                resources_dto.CLONE_G1,
                resources_dto.CLONE_R6,
            ],
            groups=[
                resources_dto.GROUP_G2,
                resources_dto.GROUP_G1,
            ],
            bundles=[
                resources_dto.BUNDLE_B1,
                resources_dto.BUNDLE_B2,
            ],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_get_multiple(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["resource", "config", "--output-format=json", "G1-clone", "R1"]
        )
        expected = ListCibResourcesDto(
            primitives=[
                resources_dto.PRIMITIVE_R1,
                resources_dto.PRIMITIVE_R2,
                resources_dto.PRIMITIVE_R3,
                resources_dto.PRIMITIVE_R4,
            ],
            clones=[resources_dto.CLONE_G1],
            groups=[resources_dto.GROUP_G1],
            bundles=[],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)


class ResourceConfigCmdMixin:
    # pylint: disable=invalid-name
    def setUp(self):
        self.new_cib_file = get_tmp_file(self._get_tmp_file_name())
        self.pcs_runner_orig = PcsRunner(
            cib_file=get_test_resource("cib-resources.xml")
        )
        self.pcs_runner_new = PcsRunner(cib_file=self.new_cib_file.name)
        write_file_to_tmpfile(
            get_test_resource("cib-empty.xml"), self.new_cib_file
        )
        self.maxDiff = None

    def tearDown(self):
        self.new_cib_file.close()

    def _get_as_json(self, runner):
        stdout, stderr, retval = runner.run(
            ["resource", "config", "--output-format=json"]
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def test_all(self):
        stdout, stderr, retval = self.pcs_runner_orig.run(
            ["resource", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        cmds = [
            split(cmd)[1:]
            for cmd in stdout.replace("\\\n", "").strip().split(";\n")
        ]
        for cmd in cmds:
            stdout, stderr, retval = self.pcs_runner_new.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )
        self.assertEqual(
            self._get_as_json(self.pcs_runner_new),
            self._get_as_json(self.pcs_runner_orig),
        )


class ResourceConfigCmd(ResourceConfigCmdMixin, TestCase):
    @staticmethod
    def _get_tmp_file_name():
        return "tier1_resource_test_config_cib.xml"



import json
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixinOld as AssertPcsMixin
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunnerOld as PcsRunner

from .common import fixture_corosync_conf_minimal


class ClusterUuidGenerateLocal(AssertPcsMixin, TestCase):
    def setUp(self):
        self.corosync_conf_file = get_tmp_file(
            "tier1_cluster_config_uuid_generate.conf"
        )
        self.pcs_runner = PcsRunner(
            cib_file=None,
            corosync_conf_opt=self.corosync_conf_file.name,
        )

    def tearDown(self):
        self.corosync_conf_file.close()

    def test_uuid_not_present(self):
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(no_cluster_uuid=True),
            self.corosync_conf_file,
        )
        self.assert_pcs_success(["cluster", "config", "uuid", "generate"])
        corosync_json_after, _ = self.pcs_runner.run(
            ["cluster", "config", "show", "--output-format=json"]
        )
        self.assertIn("cluster_uuid", json.loads(corosync_json_after))

    def test_uuid_present(self):
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(),
            self.corosync_conf_file,
        )
        self.assert_pcs_fail(
            ["cluster", "config", "uuid", "generate"],
            stdout_full="Error: Cluster UUID has already been set, use --force "
            "to override\n"
            "Error: Errors have occurred, therefore pcs is unable "
            "to continue\n",
        )

    def test_uuid_present_with_force(self):
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(),
            self.corosync_conf_file,
        )
        corosync_json_before, _ = self.pcs_runner.run(
            ["cluster", "config", "show", "--output-format=json"]
        )
        self.assert_pcs_success(
            ["cluster", "config", "uuid", "generate", "--force"],
            stdout_full="Warning: Cluster UUID has already been set\n",
        )
        corosync_json_after, _ = self.pcs_runner.run(
            ["cluster", "config", "show", "--output-format=json"]
        )
        self.assertNotEqual(
            json.loads(corosync_json_before)["cluster_uuid"],
            json.loads(corosync_json_after)["cluster_uuid"],
        )



import json
import logging
from base64 import b64encode
from unittest import (
    TestCase,
    mock,
)
from urllib.parse import urlencode

from tornado.httputil import (
    HTTPHeaders,
    HTTPServerRequest,
)
from tornado.testing import (
    AsyncTestCase,
    gen_test,
)
from tornado.web import HTTPError

from pcs.daemon import ruby_pcsd
from pcs.lib.auth.types import AuthUser

from pcs_test.tools.misc import create_patcher
from pcs_test.tools.misc import get_test_resource as rc

# Don't write errors to test output.
logging.getLogger("pcs.daemon").setLevel(logging.CRITICAL)


def create_wrapper():
    return ruby_pcsd.Wrapper(rc("/path/to/ruby_socket"))


def create_http_request():
    return HTTPServerRequest(
        method="POST",
        uri="/pcsd/uri",
        headers=HTTPHeaders({"Cookie": "cookie1=first;cookie2=second"}),
        body=str.encode(urlencode({"post-key": "post-value"})),
        host="pcsd-host:2224",
    )


patch_ruby_pcsd = create_patcher(ruby_pcsd)


class RunRuby(AsyncTestCase):
    def setUp(self):
        self.ruby_response = ""
        self.request = ruby_pcsd.RubyDaemonRequest(ruby_pcsd.SYNC_CONFIGS)
        self.wrapper = create_wrapper()
        patcher = mock.patch.object(
            self.wrapper, "send_to_ruby", self.send_to_ruby
        )
        self.addCleanup(patcher.stop)
        patcher.start()
        super().setUp()

    async def send_to_ruby(self, ruby_request):
        self.assertEqual(ruby_request, self.request)
        return self.ruby_response

    def set_run_result(self, run_result):
        self.ruby_response = json.dumps({**run_result, "logs": []})

    def assert_sinatra_result(self, result, headers, status, body):
        self.assertEqual(result.headers, headers)
        self.assertEqual(result.status, status)
        self.assertEqual(result.body, str.encode(body))

    @gen_test
    def test_correct_sending(self):
        run_result = {"next": 10}
        self.set_run_result(run_result)
        result = yield self.wrapper.run_ruby(ruby_pcsd.SYNC_CONFIGS)
        self.assertEqual(result["next"], run_result["next"])

    @gen_test
    def test_error_from_ruby(self):
        with self.assertRaises(HTTPError):
            yield self.wrapper.run_ruby(ruby_pcsd.SYNC_CONFIGS)

    @gen_test
    def test_sync_config_shortcut_success(self):
        _next = 10
        self.set_run_result({"next": _next})
        result = yield self.wrapper.sync_configs()
        self.assertEqual(result, _next)

    @patch_ruby_pcsd("now", return_value=0)
    @gen_test
    def test_sync_config_shorcut_fail(self, now):
        # pylint: disable=unused-argument
        result = yield self.wrapper.sync_configs()
        self.assertEqual(result, ruby_pcsd.DEFAULT_SYNC_CONFIG_DELAY)

    @gen_test
    def test_request(self):
        headers = {"some": "header"}
        status = 200
        body = "content"
        user = "user"
        groups = ["hacluster"]
        self.set_run_result(
            {
                "headers": headers,
                "status": status,
                "body": b64encode(str.encode(body)).decode(),
            }
        )
        http_request = create_http_request()
        self.request = ruby_pcsd.RubyDaemonRequest(
            ruby_pcsd.SINATRA,
            http_request,
            {
                "username": user,
                "groups": groups,
            },
        )
        result = yield self.wrapper.request(
            AuthUser(username=user, groups=groups), http_request
        )
        self.assert_sinatra_result(result, headers, status, body)


class ProcessResponseLog(TestCase):
    @patch_ruby_pcsd("log.from_external_source")
    @patch_ruby_pcsd("next", mock.Mock(return_value=1))
    def test_put_correct_values_to_log(self, from_external_source):
        # pylint: disable=no-self-use
        ruby_pcsd.process_response_logs(
            [
                {
                    "level": "FATAL",
                    "timestamp_usec": 1234567890,
                    "message": "ruby_message",
                }
            ]
        )
        from_external_source.assert_called_once_with(
            level=logging.CRITICAL,
            created=1234.56789,
            usecs=567890,
            message="ruby_message",
            group_id=1,
        )



from typing import List

from pcs import settings
from pcs.common import reports
from pcs.common.str_tools import split_multiline
from pcs.lib.external import CommandRunner

from .error import (
    AgentNameGuessFoundMoreThanOne,
    AgentNameGuessFoundNone,
)
from .types import (
    ResourceAgentName,
    StandardProviderTuple,
)

_IGNORED_AGENTS = frozenset(
    [
        "fence_ack_manual",
        "fence_check",
        "fence_kdump_send",
        "fence_legacy",
        "fence_na",
        "fence_node",
        "fence_nss_wrapper",
        "fence_pcmk",
        "fence_sanlockd",
        "fence_tool",
        "fence_virtd",
        "fence_vmware_helper",
    ]
)


def list_resource_agents_standards(runner: CommandRunner) -> List[str]:
    """
    Return a list of resource agents standards (ocf, lsb, ...) on the local host
    """
    # retval is the number of standards found
    stdout, dummy_stderr, dummy_retval = runner.run(
        [settings.crm_resource_binary, "--list-standards"]
    )
    return sorted(set(split_multiline(stdout)), key=str.lower)


def list_resource_agents_ocf_providers(runner: CommandRunner) -> List[str]:
    """
    Return a list of resource agents ocf providers on the local host
    """
    # retval is the number of providers found
    stdout, dummy_stderr, dummy_retval = runner.run(
        [settings.crm_resource_binary, "--list-ocf-providers"]
    )
    return sorted(set(split_multiline(stdout)), key=str.lower)


def list_resource_agents_standards_and_providers(
    runner: CommandRunner,
) -> List[StandardProviderTuple]:
    """
    Return a list of all standard[:provider] on the local host
    """
    result = []
    for standard in list_resource_agents_standards(runner):
        if standard == "ocf":
            # do not list "ocf" when we're going to list "ocf:{provider}"
            result.extend(
                [
                    StandardProviderTuple(standard, provider)
                    for provider in list_resource_agents_ocf_providers(runner)
                ]
            )
        else:
            result.append(StandardProviderTuple(standard))
    return sorted(result)


def list_resource_agents(
    runner: CommandRunner, standard_provider: StandardProviderTuple
) -> List[str]:
    """
    Return a list of resource agents of the specified standard on the local host

    standard_provider -- standard[:provider], e.g. lsb, ocf, ocf:pacemaker
    """
    # retval is 0 on success, anything else when no agents were found
    stdout, dummy_stderr, retval = runner.run(
        [
            settings.crm_resource_binary,
            "--list-agents",
            (
                f"{standard_provider.standard}:{standard_provider.provider}"
                if standard_provider.provider
                else standard_provider.standard
            ),
        ]
    )
    return (
        sorted(set(split_multiline(stdout)) - _IGNORED_AGENTS, key=str.lower)
        if retval == 0
        else []
    )


### find an agent by its name


def find_one_resource_agent_by_type(
    runner: CommandRunner,
    report_processor: reports.ReportProcessor,
    type_: str,
) -> ResourceAgentName:
    """
    Get one resource agent with the specified type from all standards:providers

    type_ -- last part of an agent's name
    """
    possible_names = _find_all_resource_agents_by_type(runner, type_)
    if len(possible_names) == 1:
        report_processor.report(
            reports.ReportItem.info(
                reports.messages.AgentNameGuessed(
                    type_, possible_names[0].full_name
                )
            )
        )
        return possible_names[0]
    if possible_names:
        raise AgentNameGuessFoundMoreThanOne(
            type_, sorted([name.full_name for name in possible_names])
        )
    raise AgentNameGuessFoundNone(type_)


def _find_all_resource_agents_by_type(
    runner: CommandRunner, type_: str
) -> List[ResourceAgentName]:
    """
    List resource agents with the specified type from all standards:providers

    type_ -- last part of an agent name
    """
    type_lower = type_.lower()
    possible_names = []
    for std_provider in list_resource_agents_standards_and_providers(runner):
        for existing_type in list_resource_agents(runner, std_provider):
            if type_lower == existing_type.lower():
                possible_names.append(
                    ResourceAgentName(
                        std_provider.standard,
                        std_provider.provider,
                        existing_type,
                    )
                )
    return possible_names



import logging
from collections import namedtuple
from typing import (
    Any,
    Dict,
)

from pcs import settings
from pcs.cli.common import middleware
from pcs.lib.commands import (
    acl,
    alert,
    booth,
    cib_options,
    cluster,
    cluster_property,
    dr,
    fencing_topology,
    node,
    pcsd,
    qdevice,
    quorum,
    remote_node,
    resource,
    resource_agent,
    sbd,
    scsi,
    services,
    status,
    stonith,
    stonith_agent,
    tag,
)
from pcs.lib.commands.constraint import colocation as constraint_colocation
from pcs.lib.commands.constraint import order as constraint_order
from pcs.lib.commands.constraint import ticket as constraint_ticket
from pcs.lib.env import LibraryEnvironment

# Note: not properly typed
_CACHE: Dict[Any, Any] = {}


def wrapper(dictionary):
    return namedtuple("wrapper", dictionary.keys())(**dictionary)


def cli_env_to_lib_env(cli_env):
    return LibraryEnvironment(
        logging.getLogger("pcs"),
        cli_env.report_processor,
        cli_env.user,
        cli_env.groups,
        cli_env.cib_data,
        cli_env.corosync_conf_data,
        booth_files_data=cli_env.booth,
        known_hosts_getter=cli_env.known_hosts_getter,
        request_timeout=cli_env.request_timeout,
    )


def lib_env_to_cli_env(lib_env, cli_env):
    if not lib_env.is_cib_live:
        cli_env.cib_data = lib_env.final_mocked_cib_content
    if not lib_env.is_corosync_conf_live:
        cli_env.corosync_conf_data = lib_env.get_corosync_conf_data()

    # TODO
    # We expect that when there is booth set up in cli_env then there is booth
    # set up in lib_env as well. The code works like that now. Once we start
    # communicate over the network, we must do extra checks in here to make
    # sure what the status really is.
    # this applies generally, not only for booth
    # corosync_conf and cib suffers with this problem as well but in this cases
    # it is dangerously hidden: when inconsistency between cli and lib
    # environment occurs, original content is put to file (which is wrong)
    if cli_env.booth:
        cli_env.booth["modified_env"] = lib_env.get_booth_env(name="").export()

    return cli_env


def bind(cli_env, run_with_middleware, run_library_command):
    def run(cli_env, *args, **kwargs):
        lib_env = cli_env_to_lib_env(cli_env)

        lib_call_result = run_library_command(lib_env, *args, **kwargs)

        # midlewares needs finish its work and they see only cli_env
        # so we need reflect some changes to cli_env
        lib_env_to_cli_env(lib_env, cli_env)

        return lib_call_result

    def decorated_run(*args, **kwargs):
        return run_with_middleware(run, cli_env, *args, **kwargs)

    return decorated_run


def bind_all(env, run_with_middleware, dictionary):
    return wrapper(
        dict(
            (exposed_fn, bind(env, run_with_middleware, library_fn))
            for exposed_fn, library_fn in dictionary.items()
        )
    )


def get_module(env, middleware_factory, name):
    if name not in _CACHE:
        _CACHE[name] = load_module(env, middleware_factory, name)
    return _CACHE[name]


def load_module(env, middleware_factory, name):
    # pylint: disable=too-many-return-statements, too-many-branches
    if name == "acl":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "create_role": acl.create_role,
                "remove_role": acl.remove_role,
                "assign_role_not_specific": acl.assign_role_not_specific,
                "assign_role_to_target": acl.assign_role_to_target,
                "assign_role_to_group": acl.assign_role_to_group,
                "unassign_role_not_specific": acl.unassign_role_not_specific,
                "unassign_role_from_target": acl.unassign_role_from_target,
                "unassign_role_from_group": acl.unassign_role_from_group,
                "create_target": acl.create_target,
                "create_group": acl.create_group,
                "remove_target": acl.remove_target,
                "remove_group": acl.remove_group,
                "add_permission": acl.add_permission,
                "remove_permission": acl.remove_permission,
                "get_config": acl.get_config,
            },
        )

    if name == "alert":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "create_alert": alert.create_alert,
                "update_alert": alert.update_alert,
                "remove_alert": alert.remove_alert,
                "add_recipient": alert.add_recipient,
                "update_recipient": alert.update_recipient,
                "remove_recipient": alert.remove_recipient,
                "get_all_alerts": alert.get_all_alerts,
            },
        )

    if name == "booth":
        bindings = {
            "config_setup": booth.config_setup,
            "config_destroy": booth.config_destroy,
            "config_text": booth.config_text,
            "config_ticket_add": booth.config_ticket_add,
            "config_ticket_remove": booth.config_ticket_remove,
            "create_in_cluster": booth.create_in_cluster,
            "remove_from_cluster": booth.remove_from_cluster,
            "restart": booth.restart,
            "config_sync": booth.config_sync,
            "enable_booth": booth.enable_booth,
            "disable_booth": booth.disable_booth,
            "start_booth": booth.start_booth,
            "stop_booth": booth.stop_booth,
            "pull_config": booth.pull_config,
            "get_status": booth.get_status,
            "ticket_grant": booth.ticket_grant,
            "ticket_revoke": booth.ticket_revoke,
        }
        if settings.booth_enable_authfile_set_enabled:
            bindings[
                "config_set_enable_authfile"
            ] = booth.config_set_enable_authfile
        if settings.booth_enable_authfile_unset_enabled:
            bindings[
                "config_unset_enable_authfile"
            ] = booth.config_unset_enable_authfile
        return bind_all(
            env,
            middleware.build(
                middleware_factory.booth_conf, middleware_factory.cib
            ),
            bindings,
        )

    if name == "cluster":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "add_link": cluster.add_link,
                "add_nodes": cluster.add_nodes,
                "corosync_authkey_change": cluster.corosync_authkey_change,
                "config_update": cluster.config_update,
                "config_update_local": cluster.config_update_local,
                "get_corosync_conf_struct": cluster.get_corosync_conf_struct,
                "node_clear": cluster.node_clear,
                "remove_links": cluster.remove_links,
                "remove_nodes": cluster.remove_nodes,
                "remove_nodes_from_cib": cluster.remove_nodes_from_cib,
                "setup": cluster.setup,
                "setup_local": cluster.setup_local,
                "update_link": cluster.update_link,
                "verify": cluster.verify,
                "generate_cluster_uuid": cluster.generate_cluster_uuid,
                "generate_cluster_uuid_local": cluster.generate_cluster_uuid_local,
            },
        )

    if name == "dr":
        return bind_all(
            env,
            middleware.build(middleware_factory.corosync_conf_existing),
            {
                "get_config": dr.get_config,
                "destroy": dr.destroy,
                "set_recovery_site": dr.set_recovery_site,
                "status_all_sites_plaintext": dr.status_all_sites_plaintext,
            },
        )

    if name == "remote_node":
        return bind_all(
            env,
            middleware.build(
                middleware_factory.cib,
                middleware_factory.corosync_conf_existing,
            ),
            {
                "node_add_remote": remote_node.node_add_remote,
                "node_add_guest": remote_node.node_add_guest,
                "node_remove_remote": remote_node.node_remove_remote,
                "node_remove_guest": remote_node.node_remove_guest,
            },
        )

    if name == "constraint_colocation":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "create_with_set": constraint_colocation.create_with_set,
                "config": constraint_colocation.config,
            },
        )

    if name == "constraint_order":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "create_with_set": constraint_order.create_with_set,
                "config": constraint_order.config,
            },
        )

    if name == "constraint_ticket":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "create_with_set": constraint_ticket.create_with_set,
                "config": constraint_ticket.config,
                "create": constraint_ticket.create,
                "remove": constraint_ticket.remove,
            },
        )

    if name == "fencing_topology":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "add_level": fencing_topology.add_level,
                "get_config": fencing_topology.get_config,
                "remove_all_levels": fencing_topology.remove_all_levels,
                "remove_levels_by_params": (
                    fencing_topology.remove_levels_by_params
                ),
                "verify": fencing_topology.verify,
            },
        )

    if name == "node":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "maintenance_unmaintenance_all": (
                    node.maintenance_unmaintenance_all
                ),
                "maintenance_unmaintenance_list": (
                    node.maintenance_unmaintenance_list
                ),
                "maintenance_unmaintenance_local": (
                    node.maintenance_unmaintenance_local
                ),
                "standby_unstandby_all": node.standby_unstandby_all,
                "standby_unstandby_list": node.standby_unstandby_list,
                "standby_unstandby_local": node.standby_unstandby_local,
            },
        )

    if name == "pcsd":
        return bind_all(
            env,
            middleware.build(),
            {"synchronize_ssl_certificate": pcsd.synchronize_ssl_certificate},
        )

    if name == "qdevice":
        return bind_all(
            env,
            middleware.build(),
            {
                "qdevice_status_text": qdevice.qdevice_status_text,
                "qdevice_setup": qdevice.qdevice_setup,
                "qdevice_destroy": qdevice.qdevice_destroy,
                "qdevice_start": qdevice.qdevice_start,
                "qdevice_stop": qdevice.qdevice_stop,
                "qdevice_kill": qdevice.qdevice_kill,
                "qdevice_enable": qdevice.qdevice_enable,
                "qdevice_disable": qdevice.qdevice_disable,
                # following commands are internal use only, called from pcsd
                "client_net_setup": qdevice.client_net_setup,
                "client_net_import_certificate": (
                    qdevice.client_net_import_certificate
                ),
                "client_net_destroy": qdevice.client_net_destroy,
                "qdevice_net_sign_certificate_request": (
                    qdevice.qdevice_net_sign_certificate_request
                ),
            },
        )

    if name == "quorum":
        return bind_all(
            env,
            middleware.build(middleware_factory.corosync_conf_existing),
            {
                "add_device": quorum.add_device,
                "get_config": quorum.get_config,
                "remove_device": quorum.remove_device,
                "remove_device_heuristics": quorum.remove_device_heuristics,
                "set_expected_votes_live": quorum.set_expected_votes_live,
                "set_options": quorum.set_options,
                "status_text": quorum.status_text,
                "status_device_text": quorum.status_device_text,
                "update_device": quorum.update_device,
                # used by ha_cluster system role
                "device_net_certificate_check_local": quorum.device_net_certificate_check_local,
                "device_net_certificate_setup_local": quorum.device_net_certificate_setup_local,
            },
        )

    if name == "resource_agent":
        return bind_all(
            env,
            middleware.build(),
            {
                "describe_agent": resource_agent.describe_agent,
                "get_agent_default_operations": resource_agent.get_agent_default_operations,
                "get_agent_metadata": resource_agent.get_agent_metadata,
                "get_agents_list": resource_agent.get_agents_list,
                "get_structured_agent_name": resource_agent.get_structured_agent_name,
                "list_agents_for_standard_and_provider": (
                    resource_agent.list_agents_for_standard_and_provider
                ),
                "list_agents": resource_agent.list_agents,
                "list_ocf_providers": resource_agent.list_ocf_providers,
                "list_standards": resource_agent.list_standards,
            },
        )

    if name == "resource":
        return bind_all(
            env,
            middleware.build(
                middleware_factory.cib,
                middleware_factory.corosync_conf_existing,
            ),
            {
                "ban": resource.ban,
                "bundle_create": resource.bundle_create,
                "bundle_reset": resource.bundle_reset,
                "bundle_update": resource.bundle_update,
                "create": resource.create,
                "create_as_clone": resource.create_as_clone,
                "create_in_group": resource.create_in_group,
                "create_into_bundle": resource.create_into_bundle,
                "disable": resource.disable,
                "disable_safe": resource.disable_safe,
                "disable_simulate": resource.disable_simulate,
                "enable": resource.enable,
                "get_configured_resources": resource.get_configured_resources,
                "get_failcounts": resource.get_failcounts,
                "group_add": resource.group_add,
                "is_any_resource_except_stonith": resource.is_any_resource_except_stonith,
                "is_any_stonith": resource.is_any_stonith,
                "manage": resource.manage,
                "move": resource.move,
                "move_autoclean": resource.move_autoclean,
                "get_resource_relations_tree": (
                    resource.get_resource_relations_tree
                ),
                "unmanage": resource.unmanage,
                "unmove_unban": resource.unmove_unban,
            },
        )

    if name == "cib_options":
        return bind_all(
            env,
            middleware.build(
                middleware_factory.cib,
            ),
            {
                "operation_defaults_config": cib_options.operation_defaults_config,
                "operation_defaults_create": cib_options.operation_defaults_create,
                "operation_defaults_remove": cib_options.operation_defaults_remove,
                "operation_defaults_update": cib_options.operation_defaults_update,
                "resource_defaults_config": cib_options.resource_defaults_config,
                "resource_defaults_create": cib_options.resource_defaults_create,
                "resource_defaults_remove": cib_options.resource_defaults_remove,
                "resource_defaults_update": cib_options.resource_defaults_update,
            },
        )

    if name == "status":
        return bind_all(
            env,
            middleware.build(
                middleware_factory.cib,
                middleware_factory.corosync_conf_existing,
            ),
            {
                "pacemaker_status_xml": status.pacemaker_status_xml,
                "full_cluster_status_plaintext": (
                    status.full_cluster_status_plaintext
                ),
            },
        )

    if name == "stonith":
        return bind_all(
            env,
            middleware.build(
                middleware_factory.cib,
                middleware_factory.corosync_conf_existing,
            ),
            {
                "create": stonith.create,
                "create_in_group": stonith.create_in_group,
                "history_get_text": stonith.history_get_text,
                "history_cleanup": stonith.history_cleanup,
                "history_update": stonith.history_update,
                "update_scsi_devices": stonith.update_scsi_devices,
                "update_scsi_devices_add_remove": stonith.update_scsi_devices_add_remove,
            },
        )

    if name == "sbd":
        return bind_all(
            env,
            middleware.build(),
            {
                "enable_sbd": sbd.enable_sbd,
                "disable_sbd": sbd.disable_sbd,
                "get_cluster_sbd_status": sbd.get_cluster_sbd_status,
                "get_cluster_sbd_config": sbd.get_cluster_sbd_config,
                "get_local_sbd_config": sbd.get_local_sbd_config,
                "initialize_block_devices": sbd.initialize_block_devices,
                "get_local_devices_info": sbd.get_local_devices_info,
                "set_message": sbd.set_message,
                "get_local_available_watchdogs": (
                    sbd.get_local_available_watchdogs
                ),
                "test_local_watchdog": sbd.test_local_watchdog,
            },
        )

    if name == "services":
        return bind_all(
            env,
            middleware.build(),
            {
                "start_service": services.start_service,
                "stop_service": services.stop_service,
                "enable_service": services.enable_service,
                "disable_service": services.disable_service,
                "get_services_info": services.get_services_info,
            },
        )
    if name == "scsi":
        return bind_all(
            env,
            middleware.build(),
            {
                "unfence_node": scsi.unfence_node,
                "unfence_node_mpath": scsi.unfence_node_mpath,
            },
        )

    if name == "stonith_agent":
        return bind_all(
            env,
            middleware.build(),
            {
                "describe_agent": stonith_agent.describe_agent,
                "list_agents": stonith_agent.list_agents,
            },
        )

    if name == "tag":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "config": tag.config,
                "create": tag.create,
                "remove": tag.remove,
                "update": tag.update,
            },
        )

    if name == "cluster_property":
        return bind_all(
            env,
            middleware.build(middleware_factory.cib),
            {
                "set_properties": cluster_property.set_properties,
            },
        )

    raise Exception("No library part '{0}'".format(name))


class Library:
    # pylint: disable=too-few-public-methods
    def __init__(self, env, middleware_factory):
        self.env = env
        self.middleware_factory = middleware_factory

    def __getattr__(self, name):
        return get_module(self.env, self.middleware_factory, name)



import time
from typing import (
    Iterable,
    Optional,
    Sequence,
    cast,
)

from pcs.common.tools import get_unique_uuid
from pcs.lib.interface.config import FacadeInterface

from .types import (
    TOKEN_ENTRY_DATETIME_FORMAT,
    TokenEntry,
)


class Facade(FacadeInterface):
    def __init__(self, parsed_config: Iterable[TokenEntry]) -> None:
        super().__init__(parsed_config)

    def _set_config(self, config: Iterable[TokenEntry]) -> None:
        super()._set_config(list(config))

    @property
    def config(self) -> Sequence[TokenEntry]:
        return tuple(cast(Iterable[TokenEntry], super().config))

    def _get_entry_by_token(self, token: str) -> Optional[TokenEntry]:
        return {entry.token: entry for entry in self.config}.get(token)

    def get_user(self, token: str) -> Optional[str]:
        entry = self._get_entry_by_token(token)
        if entry is None:
            return None
        return entry.username

    def add_entry(self, username: str, token: str) -> None:
        self._set_config(
            list(self.config)
            + [
                TokenEntry(
                    username=username,
                    token=token,
                    creation_date=time.strftime(TOKEN_ENTRY_DATETIME_FORMAT),
                )
            ]
        )

    def add_user(self, username: str) -> str:
        token = get_unique_uuid(tuple(entry.token for entry in self.config))
        self.add_entry(username, token)
        return token



from unittest import TestCase

from lxml import etree

from pcs.common.str_tools import (
    format_list,
    format_plural,
)

from pcs_test.tier0.lib.commands.test_cluster_property import ALLOWED_PROPERTIES
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import etree_to_str

property_cib = rc("cib-property.xml")
UNCHANGED_CRM_CONFIG = etree_to_str(
    etree.parse(property_cib).findall(".//crm_config")[0]
)


def get_invalid_option_messages(option_names, error=True, forceable=True):
    error_occurred = (
        "Error: Errors have occurred, therefore pcs is unable to continue\n"
    )
    use_force = ", use --force to override"
    return (
        "{severity}: invalid cluster property {option_pl} {option_name_list}, "
        "allowed options are: {allowed_properties}{use_force}\n"
        "{error_occurred}"
    ).format(
        severity="Error" if error else "Warning",
        option_name_list=format_list(option_names),
        option_pl=format_plural(option_names, "option", "options:"),
        allowed_properties=format_list(ALLOWED_PROPERTIES),
        use_force=use_force if error and forceable else "",
        error_occurred=error_occurred if error else "",
    )


class PropertyMixin(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//crm_config")[0]
        )
    )
):
    def setUp(self):
        # pylint: disable=invalid-name
        self.temp_cib = get_tmp_file("tier1_cluster_property")
        write_file_to_tmpfile(property_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        # pylint: disable=invalid-name
        self.temp_cib.close()


class TestPropertySet(PropertyMixin, TestCase):
    def test_success(self):
        self.assert_effect_single(
            (
                "property set enable-acl=true placement-strategy=utilization "
                "maintenance-mode="
            ).split(),
            """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-have-watchdog"
                        name="have-watchdog" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="HACluster"
                    />
                    <nvpair id="cib-bootstrap-options-placement-strategy"
                        name="placement-strategy" value="utilization"
                    />
                    <nvpair id="cib-bootstrap-options-enable-acl"
                        name="enable-acl" value="true"
                    />
                </cluster_property_set>
                <cluster_property_set id="second-set" score="10">
                    <nvpair id="second-set-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                </cluster_property_set>
            </crm_config>
            """,
        )

    def test_properties_to_set_missing(self):
        self.assert_pcs_fail(
            "property set".split(),
            stderr_start="\nUsage: pcs property set...",
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_missing_value(self):
        self.assert_pcs_fail(
            "property set keyword".split(),
            stderr_start="Error: missing value of 'keyword' option\n",
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unknown_properties(self):
        self.assert_pcs_fail(
            "property set unknown=value".split(),
            stderr_full=get_invalid_option_messages(["unknown"]),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unknown_properties_forced(self):
        self.assert_effect_single(
            "property set unknown=value --force".split(),
            """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-have-watchdog"
                        name="have-watchdog" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="HACluster"
                    />
                    <nvpair id="cib-bootstrap-options-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-placement-strategy"
                        name="placement-strategy" value="minimal"
                    />
                    <nvpair id="cib-bootstrap-options-enable-acl"
                        name="enable-acl" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-unknown" name="unknown"
                        value="value"
                    />
                </cluster_property_set>
                <cluster_property_set id="second-set" score="10">
                    <nvpair id="second-set-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                </cluster_property_set>
            </crm_config>
            """,
            stderr_full=get_invalid_option_messages(["unknown"], error=False),
        )

    def test_forbidden_properties(self):
        self.assert_pcs_fail(
            "property set cluster-name=NewName".split(),
            stderr_full=get_invalid_option_messages(
                ["cluster-name"], forceable=False
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_forbidden_properties_forced(self):
        self.assert_pcs_fail(
            "property set cluster-name=NewName --force".split(),
            stderr_full=get_invalid_option_messages(
                ["cluster-name"], forceable=False
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_set_stonith_watchdog_timeout(self):
        self.assert_pcs_fail(
            "property set stonith-watchdog-timeout=5s".split(),
            stderr_full=(
                "Error: stonith-watchdog-timeout can only be unset or set to 0 "
                "while SBD is disabled\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_set_stonith_watchdog_timeout_invalid_value(self):
        self.assert_pcs_fail(
            "property set stonith-watchdog-timeout=5x".split(),
            stderr_full=(
                "Error: '5x' is not a valid stonith-watchdog-timeout value, use"
                " time interval (e.g. 1, 2s, 3m, 4h, ...)\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)


class TestPropertyUnset(PropertyMixin, TestCase):
    def test_success(self):
        self.assert_effect_single(
            (
                "property unset placement-strategy enable-acl maintenance-mode"
            ).split(),
            """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair id="cib-bootstrap-options-have-watchdog"
                        name="have-watchdog" value="false"
                    />
                    <nvpair id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="HACluster"
                    />
                </cluster_property_set>
                <cluster_property_set id="second-set" score="10">
                    <nvpair id="second-set-maintenance-mode"
                        name="maintenance-mode" value="false"
                    />
                </cluster_property_set>
            </crm_config>
            """,
        )

    def test_properties_to_set_missing(self):
        self.assert_pcs_fail(
            "property unset".split(),
            stderr_start="\nUsage: pcs property unset...",
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unset_not_configured_properties(self):
        self.assert_pcs_fail(
            "property unset missing1 missing2".split(),
            stderr_full=(
                "Error: Cannot remove properties 'missing1', 'missing2', they "
                "are not present in property set 'cib-bootstrap-options', use "
                "--force to override\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(UNCHANGED_CRM_CONFIG)

    def test_unset_not_configured_properties_forced(self):
        self.assert_effect_single(
            "property unset missing1 missing2 --force".split(),
            UNCHANGED_CRM_CONFIG,
            stderr_full=(
                "Warning: Cannot remove properties 'missing1', 'missing2', they "
                "are not present in property set 'cib-bootstrap-options'\n"
            ),
        )



from typing import (
    Collection,
    Mapping,
    Union,
)

from pcs.common import reports
from pcs.common.types import StringSequence
from pcs.lib import cluster_property
from pcs.lib.cib import (
    nvpair_multi,
    rule,
)
from pcs.lib.cib.tools import IdProvider
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.resource_agent import (
    ResourceAgentError,
    ResourceAgentFacade,
    ResourceAgentMetadata,
)
from pcs.lib.resource_agent import const as ra_const
from pcs.lib.resource_agent import resource_agent_error_to_report_item
from pcs.lib.resource_agent.facade import ResourceAgentFacadeFactory


def _get_property_facade_list(
    report_processor: reports.ReportProcessor,
    factory: ResourceAgentFacadeFactory,
) -> list[ResourceAgentFacade]:
    pacemaker_daemons = [
        ra_const.PACEMAKER_BASED,
        ra_const.PACEMAKER_CONTROLD,
        ra_const.PACEMAKER_SCHEDULERD,
    ]
    cluster_property_facade_list = []
    for daemon in pacemaker_daemons:
        try:
            cluster_property_facade_list.append(
                factory.facade_from_pacemaker_daemon_name(daemon)
            )
        except ResourceAgentError as e:
            report_processor.report_list(
                [
                    resource_agent_error_to_report_item(
                        e, reports.ReportItemSeverity.error()
                    )
                ]
            )
    if report_processor.has_errors:
        raise LibraryError()
    return cluster_property_facade_list


# backward compatibility layer - export cluster property metadata in the legacy
# format
def _cluster_property_metadata_to_dict(
    metadata: ResourceAgentMetadata,
) -> dict[str, dict[str, Union[bool, str, StringSequence]]]:
    banned_props = ["dc-version", "cluster-infrastructure"]
    basic_props = {
        "batch-limit": "Batch Limit",
        "no-quorum-policy": "No Quorum Policy",
        "symmetric-cluster": "Symmetric",
        "stonith-enabled": "Stonith Enabled",
        "stonith-action": "Stonith Action",
        "cluster-delay": "Cluster Delay",
        "stop-orphan-resources": "Stop Orphan Resources",
        "stop-orphan-actions": "Stop Orphan Actions",
        "start-failure-is-fatal": "Start Failure is Fatal",
        "pe-error-series-max": "PE Error Storage",
        "pe-warn-series-max": "PE Warning Storage",
        "pe-input-series-max": "PE Input Storage",
        "enable-acl": "Enable ACLs",
    }
    property_definition = {}
    for parameter in metadata.parameters:
        if parameter.name in banned_props:
            continue
        single_property_dict: dict[str, Union[bool, str, StringSequence]] = {
            "name": parameter.name,
            "shortdesc": parameter.shortdesc or "",
            "longdesc": parameter.longdesc or "",
            "type": parameter.type,
            "default": parameter.default or "",
            "advanced": parameter.name not in basic_props or parameter.advanced,
            "readable_name": basic_props.get(parameter.name, parameter.name),
            "source": metadata.name.type,
        }
        if parameter.enum_values is not None:
            single_property_dict["enum"] = parameter.enum_values
            single_property_dict["type"] = "enum"
        property_definition[parameter.name] = single_property_dict
    return property_definition


def get_cluster_properties_definition_legacy(
    env: LibraryEnvironment,
) -> dict[str, dict[str, Union[bool, str, StringSequence]]]:
    """
    Return cluster properties definition in the legacy dictionary format.

    env -- provides communication with externals
    """
    facade_factory = ResourceAgentFacadeFactory(
        env.cmd_runner(), env.report_processor
    )
    property_dict = {}
    for facade in _get_property_facade_list(
        env.report_processor, facade_factory
    ):
        property_dict.update(
            _cluster_property_metadata_to_dict(facade.metadata)
        )
    return property_dict


def set_properties(
    env: LibraryEnvironment,
    cluster_properties: Mapping[str, str],
    force_flags: Collection[reports.types.ForceCode] = (),
) -> None:
    """
    Set specified pacemaker cluster properties, remove those with empty values.

    env -- provides communication with externals
    cluster_properties -- dictionary of cluster property names and values
    force_flags -- list of flags codes
    """
    cib = env.get_cib()
    id_provider = IdProvider(cib)
    force = reports.codes.FORCE in force_flags
    cluster_property_set_el = (
        cluster_property.get_cluster_property_set_element_legacy(
            cib, id_provider
        )
    )
    set_id = cluster_property_set_el.get("id", "")

    property_facade_list = _get_property_facade_list(
        env.report_processor,
        ResourceAgentFacadeFactory(env.cmd_runner(), env.report_processor),
    )

    configured_properties = [
        nvpair_dto.name
        for nvpair_dto in nvpair_multi.nvset_element_to_dto(
            cluster_property_set_el,
            rule.RuleInEffectEvalDummy(),
        ).nvpairs
    ]

    env.report_processor.report_list(
        cluster_property.validate_set_cluster_properties(
            property_facade_list,
            set_id,
            configured_properties,
            cluster_properties,
            env.service_manager,
            force=force,
        )
    )

    if env.report_processor.has_errors:
        raise LibraryError()

    nvpair_multi.nvset_update(
        cluster_property_set_el,
        id_provider,
        cluster_properties,
    )
    env.push_cib()



import os.path
from textwrap import dedent

from pcs import settings

from pcs_test.tools.command_env.mock_runner import Call as RunnerCall


class CorosyncShortcuts:
    qdevice_generated_cert_path = os.path.join(
        settings.corosync_qdevice_net_client_certs_dir,
        "qdevice-net-node.crq",
    )
    qdevice_pk12_cert_path = os.path.join(
        settings.corosync_qdevice_net_client_certs_dir,
        "qdevice-net-node.p12",
    )

    def __init__(self, calls):
        self.__calls = calls

    def version(
        self,
        name="runner.corosync.version",
        version="2.4.0",
        instead=None,
        before=None,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync", "-v"],
                stdout=dedent(
                    f"""\
                    Corosync Cluster Engine, version '{version}'
                    Copyright...
                    """
                ),
            ),
            before=before,
            instead=instead,
        )

    def qdevice_init_cert_storage(
        self,
        ca_file_path,
        stdout="",
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_init_cert_storage",
    ):
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-qdevice-net-certutil", "-i", "-c", ca_file_path],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def qdevice_generate_cert(
        self,
        cluster_name,
        cert_req_path=None,
        stdout=None,
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_generate_cert",
    ):
        if stdout is not None and cert_req_path is not None:
            raise AssertionError(
                "Cannot specify both 'cert_req_path' and 'stdout'"
            )
        if stdout is None and cert_req_path is None:
            cert_req_path = self.qdevice_generated_cert_path
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-qdevice-net-certutil", "-r", "-n", cluster_name],
                stdout=(
                    stdout
                    if stdout is not None
                    else f"Certificate request stored in {cert_req_path}\n"
                ),
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def qdevice_get_pk12(
        self,
        cert_path="cert path",
        output_path=None,
        stdout=None,
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_get_pk12",
    ):
        if stdout is not None and output_path is not None:
            raise AssertionError(
                "Cannot specify both 'output_path' and 'stdout'"
            )
        if stdout is None and output_path is None:
            output_path = self.qdevice_pk12_cert_path
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-qdevice-net-certutil", "-M", "-c", cert_path],
                stdout=(
                    stdout
                    if stdout is not None
                    else f"Certificate stored in {output_path}\n"
                ),
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def qdevice_import_pk12(
        self,
        pk12_file_path,
        stdout="",
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_import_pk12",
    ):
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-qdevice-net-certutil", "-m", "-c", pk12_file_path],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def qdevice_list_certs(
        self,
        stdout=None,
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_list_certs",
    ):
        if stdout is None:
            stdout = dedent(
                """\
                Certificate Nickname   Trust Attributes
                                       SSL,S/MIME,JAR/XPI

                QNet CA                CT,c,c
                Cluster Cert           u,u,u
                """
            )
        self.__calls.place(
            name,
            RunnerCall(
                [
                    settings.certutil_executable,
                    "-d",
                    settings.corosync_qdevice_net_client_certs_dir,
                    "-L",
                ],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def qdevice_show_cert(
        self,
        cert_name,
        stdout,
        ascii_only=False,
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_show_cert",
    ):
        cmd = [
            settings.certutil_executable,
            "-d",
            settings.corosync_qdevice_net_client_certs_dir,
            "-L",
            "-n",
            cert_name,
        ]
        if ascii_only:
            cmd.append("-a")
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def quorum_status(
        self,
        node_list=None,
        stdout=None,
        stderr="",
        returncode=0,
        name="runner.corosync.quorum_status",
    ):
        if bool(node_list) == bool(stdout):
            raise AssertionError(
                "Exactly one of 'node_list', 'stdout' must be specified"
            )
        if node_list:
            stdout = dedent(
                """\
                Quorum information
                ------------------
                Date:             Fri Jan 16 13:03:28 2015
                Quorum provider:  corosync_votequorum
                Nodes:            {nodes_num}
                Node ID:          1
                Ring ID:          19860
                Quorate:          Yes\n
                Votequorum information
                ----------------------
                Expected votes:   {nodes_num}
                Highest expected: {nodes_num}
                Total votes:      {nodes_num}
                Quorum:           {quorum_num}
                Flags:            Quorate\n
                Membership information
                ----------------------
                    Nodeid      Votes    Qdevice Name
                {nodes}\
                """
            ).format(
                nodes_num=len(node_list),
                quorum_num=(len(node_list) // 2) + 1,
                nodes="".join(
                    [
                        _quorum_status_node_fixture(node_id, node)
                        for node_id, node in enumerate(node_list, 1)
                    ]
                ),
            )
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-quorumtool", "-p"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )


def _quorum_status_node_fixture(node_id, node_name, votes=1, is_local=False):
    _local = " (local)" if is_local else ""
    return (
        f"         {node_id}          {votes}         NR {node_name}{_local}\n"
    )



import dataclasses
import signal
from datetime import timedelta
from logging import Logger
from multiprocessing import Process
from multiprocessing.pool import worker as mp_worker_init  # type: ignore
from queue import Queue
from unittest import mock

from tornado.testing import gen_test

from pcs import settings
from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
)
from pcs.common.async_tasks.types import (
    TaskFinishType,
    TaskKillReason,
)
from pcs.daemon.async_tasks.scheduler import TaskNotFoundError
from pcs.daemon.async_tasks.types import Command
from pcs.daemon.async_tasks.worker import executor
from pcs.daemon.async_tasks.worker.report_processor import WorkerReportProcessor
from pcs.daemon.async_tasks.worker.types import (
    Message,
    TaskExecuted,
    TaskFinished,
)

from .dummy_commands import (
    RESULT,
    test_command_map,
)
from .helpers import (
    AUTH_USER,
    DATETIME_NOW,
    AssertTaskStatesMixin,
    MockDateTimeNowMixin,
    MockOsKillMixin,
    PermissionsCheckerMock,
    SchedulerBaseAsyncTestCase,
)

COMMAND_OPTIONS = CommandOptionsDto(request_timeout=None)
executor.worker_com = Queue()  # patched at runtime


class IntegrationBaseTestCase(SchedulerBaseAsyncTestCase):
    async def perform_actions(self, message_count):
        # pylint: disable=protected-access
        """
        USE THIS FUNCTION IN TIER0 TESTS instead of the scheduler function with
        the same name to guarantee consistency between test runs. This function
        guarantees that worker_com is emptied in one call, removing variability
        between test runs.
        """
        # TODO: remove this function
        del message_count
        await self.scheduler.perform_actions()

    def execute_tasks(self, task_ident_list):
        """Simulates process pool workers launching tasks

        Emits one message into worker_com queue per task. Process IDs of these
        tasks are task_idents stripped of the "id" prefix.
        :param task_ident_list: Contains task_idents of tasks to execute
        """
        for task_ident in task_ident_list:
            self.worker_com.put_nowait(
                Message(task_ident, TaskExecuted(int(task_ident[2:])))
            )

    def finish_tasks(
        self, task_ident_list, finish_type=TaskFinishType.SUCCESS, result=None
    ):
        """Simulates process pool workers handing over results when task ends

        Emits one message into worker_com queue per task
        :param task_ident_list: Contains task_idents of tasks that finished
        :param finish_type: Task finish type for all task_idents
        :param result: Return value of an executed function for all task_idents
        """
        for task_ident in task_ident_list:
            self.worker_com.put_nowait(
                Message(task_ident, TaskFinished(finish_type, result))
            )


class StateChangeTest(AssertTaskStatesMixin, IntegrationBaseTestCase):
    """Tests that check that task state is correct at each stage

    Task state changes prompted by errors or other special events are tested
    in the specific classes, these are just baseline tests for how everything
    should work in an ideal (error-free) scenario.
    """

    @gen_test
    async def test_created_from_empty(self):
        self._create_tasks(5)
        self.assert_task_state_counts_equal(5, 0, 0, 0)

    @gen_test
    async def test_created_on_top_of_existing(self):
        self._create_tasks(5)
        await self.perform_actions(0)
        self.execute_tasks(["id0", "id1", "id2"])
        await self.perform_actions(3)
        self._create_tasks(2, start_from=5)
        # 3/5 were executed, remaining are queued, 2 new arrived
        self.assert_task_state_counts_equal(2, 2, 3, 0)

    @gen_test
    async def test_created_to_scheduled(self):
        self._create_tasks(4)
        await self.perform_actions(0)
        self.assert_task_state_counts_equal(0, 4, 0, 0)

    @gen_test
    async def test_scheduled_to_executed(self):
        self._create_tasks(4)
        await self.perform_actions(0)
        # Tasks are scheduled, now 2 will start executing
        self.execute_tasks(["id0", "id1"])
        await self.perform_actions(2)
        self.assert_task_state_counts_equal(0, 2, 2, 0)

    @gen_test
    async def test_executed_to_finished(self):
        self._create_tasks(1)
        await self.perform_actions(0)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        self.finish_tasks(["id0"])
        await self.perform_actions(1)
        self.assert_task_state_counts_equal(0, 0, 0, 1)


class GarbageCollectionTimeoutTests(
    MockOsKillMixin, MockDateTimeNowMixin, IntegrationBaseTestCase
):
    """Testing garbage collection after timeouts

    Testing garbage collection for events caused by the scheduler like timeouts.
    """

    def setUp(self):
        super().setUp()
        # This mock already sets current time to DATETIME_NOW
        self.mock_datetime_now = self._init_mock_datetime_now()
        self.mock_os_kill = self._init_mock_os_kill()

    async def _run_gc_and_assert_state(
        self, timeout_s, task_finish_type, task_kill_reason
    ):
        self.mock_datetime_now.return_value = DATETIME_NOW + timedelta(
            seconds=timeout_s + 1
        )
        await self.perform_actions(0)
        task_info = self.scheduler.get_task("id0", AUTH_USER)
        self.assertEqual(task_finish_type, task_info.task_finish_type)
        self.assertEqual(task_kill_reason, task_info.kill_reason)

    @gen_test
    async def test_get_task_removes_finished(self):
        self._create_tasks(1)
        await self.perform_actions(0)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        self.finish_tasks(["id0"])
        await self.perform_actions(1)
        self.scheduler.get_task("id0", AUTH_USER)
        # pylint: disable=protected-access
        self.assertIsNotNone(
            self.scheduler._task_register["id0"]._to_delete_timestamp
        )

    @gen_test
    async def test_created_defunct_timeout(self):
        # Nothing should happen, created tasks can't be defunct
        self._create_tasks(1)
        await self._run_gc_and_assert_state(
            settings.task_unresponsive_timeout_seconds,
            TaskFinishType.UNFINISHED,
            task_kill_reason=None,
        )

    @gen_test
    async def test_scheduled_defunct_timeout(self):
        # Nothing should happen, scheduled tasks can't be defunct
        self._create_tasks(1)
        await self.perform_actions(0)
        await self._run_gc_and_assert_state(
            settings.task_unresponsive_timeout_seconds,
            TaskFinishType.UNFINISHED,
            task_kill_reason=None,
        )

    @gen_test
    async def test_executed_defunct_timeout(self):
        # Task should be killed
        self._create_tasks(1)
        await self.perform_actions(0)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        await self._run_gc_and_assert_state(
            settings.task_unresponsive_timeout_seconds,
            TaskFinishType.KILL,
            TaskKillReason.COMPLETION_TIMEOUT,
        )

    @gen_test
    async def test_finished_defunct_timeout(self):
        # Only tasks in EXECUTED state can become defunct. In this case,
        # task is going to become ABANDONED and is deleted
        self._create_tasks(2)
        await self.perform_actions(0)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        await self._run_gc_and_assert_state(
            settings.task_unresponsive_timeout_seconds,
            TaskFinishType.KILL,
            TaskKillReason.COMPLETION_TIMEOUT,
        )
        await self.perform_actions(0)
        with self.assertRaises(TaskNotFoundError):
            self.scheduler.get_task("id0", AUTH_USER)
        # If the guard task was removed, this fails the test case
        self.scheduler.get_task("id1", AUTH_USER)

    @gen_test
    async def test_created_abandoned_timeout(self):
        self._create_tasks(1)
        await self._run_gc_and_assert_state(
            settings.task_abandoned_timeout_seconds,
            TaskFinishType.UNFINISHED,
            task_kill_reason=None,
        )

    @gen_test
    async def test_scheduled_abandoned_timeout(self):
        self._create_tasks(1)
        await self.perform_actions(0)
        await self._run_gc_and_assert_state(
            settings.task_abandoned_timeout_seconds,
            TaskFinishType.UNFINISHED,
            task_kill_reason=None,
        )

    @gen_test
    async def test_executed_abandoned_timeout(self):
        self._create_tasks(1)
        await self.perform_actions(0)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        await self._run_gc_and_assert_state(
            settings.task_abandoned_timeout_seconds,
            TaskFinishType.UNFINISHED,
            task_kill_reason=None,
        )

    @gen_test
    async def test_finished_abandoned_timeout(self):
        self._create_tasks(1)
        await self.perform_actions(0)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        self.finish_tasks(["id0"])
        await self.perform_actions(1)
        self.mock_datetime_now.return_value = DATETIME_NOW + timedelta(
            seconds=settings.task_abandoned_timeout_seconds + 1
        )
        # Garbage collector deletes an abandoned task right away
        await self.perform_actions(0)
        with self.assertRaises(TaskNotFoundError):
            self.scheduler.get_task("id0", AUTH_USER)


class GarbageCollectionUserKillTests(
    MockOsKillMixin, AssertTaskStatesMixin, IntegrationBaseTestCase
):
    """Testing garbage collection after user kills a task

    Tests the garbage collection after user intervention to running tasks.
    """

    def setUp(self):
        super().setUp()
        self.mock_os_kill = self._init_mock_os_kill()

    def assert_end_state(self):
        task_info_killed = self.scheduler.get_task("id0", AUTH_USER)
        self.assertEqual(TaskFinishType.KILL, task_info_killed.task_finish_type)
        self.assertEqual(TaskKillReason.USER, task_info_killed.kill_reason)

        task_info_alive = self.scheduler.get_task("id1", AUTH_USER)
        self.assertEqual(
            TaskFinishType.UNFINISHED, task_info_alive.task_finish_type
        )
        self.assertIsNone(task_info_alive.kill_reason)

    @gen_test
    async def test_kill_created(self):
        self._create_tasks(2)
        self.scheduler.kill_task("id0", AUTH_USER)
        # Kill_task doesn't produce any messages since the worker is killed by
        # the system
        await self.perform_actions(0)
        self.assert_task_state_counts_equal(0, 1, 0, 1)

        self.mock_os_kill.assert_not_called()
        self.assert_end_state()

    @gen_test
    async def test_kill_scheduled(self):
        self._create_tasks(2)
        await self.perform_actions(0)
        self.scheduler.kill_task("id0", AUTH_USER)
        # Garbage collection waits until the task is executed and then kills
        # the worker
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        await self.perform_actions(0)
        self.assert_task_state_counts_equal(0, 1, 0, 1)

        self.mock_os_kill.assert_called_once()
        self.assert_end_state()

    @gen_test
    async def test_kill_executed(self):
        self._create_tasks(2)
        await self.perform_actions(0)
        self.execute_tasks(["id0", "id1"])
        await self.perform_actions(2)
        self.scheduler.kill_task("id0", AUTH_USER)
        await self.perform_actions(0)
        self.assert_task_state_counts_equal(0, 0, 1, 1)

        self.mock_os_kill.assert_called_once()
        self.assert_end_state()

    @gen_test
    async def test_kill_finished(self):
        self._create_tasks(2)
        await self.perform_actions(0)
        self.execute_tasks(["id0", "id1"])
        await self.perform_actions(2)
        self.finish_tasks(["id0"])
        await self.perform_actions(1)
        # When scheduler picks up finished tasks, it sends a signal to worker
        # to resume via os.kill
        self.mock_os_kill.reset_mock()
        self.scheduler.kill_task("id0", AUTH_USER)
        await self.perform_actions(0)
        self.assert_task_state_counts_equal(0, 0, 1, 1)

        self.mock_os_kill.assert_not_called()

        task_info_not_killed = self.scheduler.get_task("id0", AUTH_USER)
        self.assertEqual(
            TaskFinishType.SUCCESS, task_info_not_killed.task_finish_type
        )
        self.assertEqual(TaskKillReason.USER, task_info_not_killed.kill_reason)

        task_info_alive = self.scheduler.get_task("id1", AUTH_USER)
        self.assertEqual(
            TaskFinishType.UNFINISHED, task_info_alive.task_finish_type
        )
        self.assertIsNone(task_info_alive.kill_reason)


class TaskResultsTests(MockOsKillMixin, IntegrationBaseTestCase):
    """These tests check all task outcomes with real task_executor

    These test go one level deeper to include task_executor and test its
    behavior. All possible task outcomes are tested.
    """

    # pylint: disable=protected-access

    def setUp(self):
        super().setUp()
        self.addCleanup(mock.patch.stopall)
        mock.patch(
            "pcs.daemon.async_tasks.worker.executor.COMMAND_MAP",
            test_command_map,
        ).start()
        mock.patch(
            "pcs.daemon.async_tasks.worker.executor.getLogger", spec=Logger
        ).start()
        mock.patch(
            "pcs.daemon.async_tasks.worker.executor.worker_com",
            self.worker_com,
        ).start()
        lib_env_mock = (
            mock.patch(
                "pcs.daemon.async_tasks.worker.executor.LibraryEnvironment"
            )
            .start()
            .return_value
        )
        mock.patch(
            "pcs.daemon.async_tasks.worker.executor.PermissionsChecker",
            lambda _: PermissionsCheckerMock({}),
        ).start()
        lib_env_mock.report_processor = WorkerReportProcessor(
            self.worker_com, "id0"
        )
        # Os.kill is used to pause the worker and we do not want to pause tests
        self._init_mock_os_kill()

    def _new_task(self, task_id, cmd):
        with mock.patch(
            "pcs.daemon.async_tasks.scheduler.get_unique_uuid"
        ) as mock_uuid:
            mock_uuid.return_value = task_id
            self.scheduler.new_task(
                Command(CommandDto(cmd, {}, COMMAND_OPTIONS)),
                AUTH_USER,
            )

    @gen_test
    async def test_task_successful_no_result_with_reports(self):
        # How is no result different from a None return value in DTO?
        # Functions without return values also return None - should we
        # distinguish between cases of ex/implicitly returned None
        task_id = "id0"
        self._new_task(task_id, "success_with_reports")
        await self.perform_actions(0)
        # This task sends one report and returns immediately, task_executor
        # sends two messages - TaskExecuted and TaskFinished
        executor.task_executor(
            self.scheduler._task_register[task_id].to_worker_command()
        )
        await self.perform_actions(3)

        task_info = self.scheduler.get_task(task_id, AUTH_USER)
        self.assertEqual(1, len(task_info.reports))
        self.assertEqual(TaskFinishType.SUCCESS, task_info.task_finish_type)
        self.assertIsNone(task_info.result)

    @gen_test
    async def test_task_successful_with_result(self):
        task_id = "id0"
        self._new_task(task_id, "success")
        await self.perform_actions(0)
        # This task sends no reports and returns immediately, task_executor
        # sends two messages - TaskExecuted and TaskFinished
        executor.task_executor(
            self.scheduler._task_register[task_id].to_worker_command()
        )
        await self.perform_actions(2)

        task_info = self.scheduler.get_task(task_id, AUTH_USER)
        self.assertEqual(0, len(task_info.reports))
        self.assertEqual(TaskFinishType.SUCCESS, task_info.task_finish_type)
        self.assertEqual(RESULT, task_info.result)

    @gen_test
    async def test_task_error(self):
        task_id = "id0"
        self._new_task(task_id, "lib_exc")
        await self.perform_actions(0)
        # This task immediately raises a LibraryException and executor detects
        # that as an error, sends two messages - TaskExecuted and TaskFinished
        executor.task_executor(
            self.scheduler._task_register[task_id].to_worker_command()
        )
        await self.perform_actions(2)

        task_info = self.scheduler.get_task(task_id, AUTH_USER)
        self.assertEqual(0, len(task_info.reports))
        self.assertEqual(TaskFinishType.FAIL, task_info.task_finish_type)
        self.assertIsNone(task_info.result)

    @gen_test
    async def test_task_unhandled_exception(self):
        task_id = "id0"
        self._new_task(task_id, "unhandled_exc")
        await self.perform_actions(0)
        # This task immediately raises an Exception which the executor catches
        # and logs accordingly
        executor.task_executor(
            self.scheduler._task_register[task_id].to_worker_command()
        )
        await self.perform_actions(2)

        task_info = self.scheduler.get_task(task_id, AUTH_USER)
        self.assertEqual(0, len(task_info.reports))
        self.assertEqual(
            TaskFinishType.UNHANDLED_EXCEPTION, task_info.task_finish_type
        )
        self.assertIsNone(task_info.result)

    @gen_test
    async def test_wait_for_task(self):
        task_id = "id0"
        self._new_task(task_id, "success")
        await self.perform_actions(0)
        executor.task_executor(
            self.scheduler._task_register[task_id].to_worker_command()
        )
        await self.perform_actions(2)

        task_info = await self.scheduler.wait_for_task(task_id, AUTH_USER)
        self.assertEqual(0, len(task_info.reports))
        self.assertEqual(TaskFinishType.SUCCESS, task_info.task_finish_type)
        self.assertEqual(RESULT, task_info.result)


@mock.patch("pcs.daemon.async_tasks.task.os.kill")
class DeadlockTests(
    MockOsKillMixin, AssertTaskStatesMixin, IntegrationBaseTestCase
):
    # pylint: disable=protected-access

    def setUp(self):
        super().setUp()
        self.addCleanup(mock.patch.stopall)
        self.scheduler._config = dataclasses.replace(
            self.scheduler._config, deadlock_threshold_timeout=0
        )
        self.process_cls_mock = mock.Mock()
        self.process_obj_mock = mock.Mock(spec=Process)
        self.process_cls_mock.return_value = self.process_obj_mock
        mock.patch(
            "pcs.daemon.async_tasks.scheduler.mp.Process", self.process_cls_mock
        ).start()

    @gen_test
    async def test_deadlock_mitigation(self, mock_kill):
        self._create_tasks(2)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        # deadlock detected, new tmp worker spawned
        self.assert_task_state_counts_equal(0, 1, 1, 0)
        self.process_cls_mock.assert_called_once_with(
            group=None,
            target=mp_worker_init,
            args=(
                self.mp_pool_mock._inqueue,
                self.mp_pool_mock._outqueue,
                executor.worker_init,
                (self.worker_com, self.logging_queue),
                1,
                False,
            ),
        )
        self.process_obj_mock.start.assert_called_once_with()
        self.process_obj_mock.close.assert_not_called()
        self.execute_tasks(["id1"])
        self.process_obj_mock.is_alive.return_value = True
        await self.perform_actions(0)
        # tmp worker started executing a task
        self.assert_task_state_counts_equal(0, 0, 2, 0)
        self.process_obj_mock.close.assert_not_called()
        self.finish_tasks(["id1"])
        self.process_obj_mock.is_alive.return_value = False
        mock_kill.assert_not_called()
        await self.perform_actions(1)
        mock_kill.assert_called_once_with(1, signal.SIGCONT)
        # tmp worker finished the task and terminated itself
        self.assert_task_state_counts_equal(0, 0, 1, 1)
        self.process_obj_mock.close.assert_called_once_with()

    @gen_test
    async def test_max_worker_count_reached(self, mock_kill):
        self.scheduler._config = dataclasses.replace(
            self.scheduler._config, max_worker_count=1
        )
        self._create_tasks(3)
        self.execute_tasks(["id0"])
        await self.perform_actions(1)
        self.assert_task_state_counts_equal(0, 2, 1, 0)
        self.process_cls_mock.assert_not_called()
        self.process_obj_mock.assert_not_called()
        mock_kill.assert_not_called()



from functools import partial
from os.path import join as join_path
from ssl import OP_NO_SSLv2
from unittest import TestCase

from pcs import settings
from pcs.daemon import env

from pcs_test.tools.misc import create_setup_patch_mixin


class Logger:
    def __init__(self):
        self.errors = []
        self.warnings = []

    def error(self, error):
        self.errors.append(error)

    def warning(self, warning):
        self.warnings.append(warning)


class Prepare(TestCase, create_setup_patch_mixin(env)):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.path_exists = self.setup_patch("path_exists", return_value=True)
        self.logger = Logger()
        self.maxDiff = None

    def assert_environ_produces_modified_pcsd_env(
        self, environ=None, specific_env_values=None, errors=None, warnings=None
    ):
        pcsd_dir = partial(join_path, settings.pcsd_exec_location)
        default_env_values = {
            env.PCSD_PORT: settings.pcsd_default_port,
            env.PCSD_SSL_CIPHERS: settings.default_ssl_ciphers,
            env.PCSD_SSL_OPTIONS: env.str_to_ssl_options(
                settings.default_ssl_options, []
            ),
            env.PCSD_BIND_ADDR: {None},
            env.NOTIFY_SOCKET: None,
            env.PCSD_DEBUG: False,
            env.PCSD_DISABLE_GUI: False,
            env.PCSD_SESSION_LIFETIME: settings.gui_session_lifetime_seconds,
            env.PCSD_STATIC_FILES_DIR: pcsd_dir(env.PCSD_STATIC_FILES_DIR_NAME),
            env.PCSD_DEV: False,
            env.PCSD_WORKER_COUNT: settings.pcsd_worker_count,
            env.PCSD_WORKER_RESET_LIMIT: settings.pcsd_worker_reset_limit,
            env.PCSD_MAX_WORKER_COUNT: settings.pcsd_worker_count
            + settings.pcsd_temporary_workers,
            env.PCSD_DEADLOCK_THRESHOLD_TIMEOUT: settings.pcsd_deadlock_threshold_timeout,
            env.PCSD_CHECK_INTERVAL_MS: settings.async_api_scheduler_interval_ms,
            env.PCSD_TASK_ABANDONED_TIMEOUT: settings.task_abandoned_timeout_seconds,
            env.PCSD_TASK_UNRESPONSIVE_TIMEOUT: settings.task_unresponsive_timeout_seconds,
            env.PCSD_TASK_DELETION_TIMEOUT: settings.task_deletion_timeout_seconds,
            "has_errors": False,
        }
        if specific_env_values is None:
            specific_env_values = {}

        # compare as dict because of clearer error report
        self.assertEqual(
            dict(env.prepare_env(environ or {}, self.logger)._asdict()),
            {**default_env_values, **specific_env_values},
        )
        self.assertEqual(self.logger.errors, errors or [])
        self.assertEqual(self.logger.warnings, warnings or [])

    def test_nothing_in_environ(self):
        self.assert_environ_produces_modified_pcsd_env()

    def test_many_valid_environment_changes(self):
        pcsd_dir = partial(join_path, env.PCSD_LOCAL_DIR)
        session_lifetime = 10
        environ = {
            env.PCSD_PORT: "1234",
            env.PCSD_SSL_CIPHERS: "DEFAULT:!3DES:@STRENGTH",
            env.PCSD_SSL_OPTIONS: "OP_NO_SSLv2",
            env.PCSD_BIND_ADDR: "abc",
            env.NOTIFY_SOCKET: "xyz",
            env.PCSD_DEBUG: "true",
            env.PCSD_DISABLE_GUI: "true",
            env.PCSD_SESSION_LIFETIME: str(session_lifetime),
            env.PCSD_DEV: "true",
            env.PCSD_WORKER_COUNT: "1",
            env.PCSD_WORKER_RESET_LIMIT: "2",
            env.PCSD_MAX_WORKER_COUNT: "3",
            env.PCSD_DEADLOCK_THRESHOLD_TIMEOUT: "4",
            env.PCSD_CHECK_INTERVAL_MS: "5",
            env.PCSD_TASK_ABANDONED_TIMEOUT: "6",
            env.PCSD_TASK_UNRESPONSIVE_TIMEOUT: "7",
            env.PCSD_TASK_DELETION_TIMEOUT: "8",
        }
        self.assert_environ_produces_modified_pcsd_env(
            environ=environ,
            specific_env_values={
                env.PCSD_PORT: environ[env.PCSD_PORT],
                env.PCSD_SSL_CIPHERS: environ[env.PCSD_SSL_CIPHERS],
                env.PCSD_SSL_OPTIONS: OP_NO_SSLv2,
                env.PCSD_BIND_ADDR: {environ[env.PCSD_BIND_ADDR]},
                env.NOTIFY_SOCKET: environ[env.NOTIFY_SOCKET],
                env.PCSD_DEBUG: True,
                env.PCSD_DISABLE_GUI: True,
                env.PCSD_SESSION_LIFETIME: session_lifetime,
                env.PCSD_STATIC_FILES_DIR: pcsd_dir(
                    env.PCSD_STATIC_FILES_DIR_NAME
                ),
                env.PCSD_DEV: True,
                env.PCSD_WORKER_COUNT: 1,
                env.PCSD_WORKER_RESET_LIMIT: 2,
                env.PCSD_MAX_WORKER_COUNT: 3,
                env.PCSD_DEADLOCK_THRESHOLD_TIMEOUT: 4,
                env.PCSD_CHECK_INTERVAL_MS: 5,
                env.PCSD_TASK_ABANDONED_TIMEOUT: 6,
                env.PCSD_TASK_UNRESPONSIVE_TIMEOUT: 7,
                env.PCSD_TASK_DELETION_TIMEOUT: 8,
            },
        )

    def test_error_on_noninteger_session_lifetime(self):
        environ = {env.PCSD_SESSION_LIFETIME: "invalid"}
        self.assert_environ_produces_modified_pcsd_env(
            environ,
            specific_env_values={**environ, "has_errors": True},
            errors=[
                "Invalid PCSD_SESSION_LIFETIME value 'invalid'"
                " (it must be an integer)"
            ],
        )

    def test_report_invalid_ssl_ciphers(self):
        environ = {env.PCSD_SSL_CIPHERS: "invalid ;@{}+ ciphers"}
        self.assert_environ_produces_modified_pcsd_env(
            environ,
            specific_env_values={**environ, "has_errors": True},
            errors=["Invalid ciphers: '('No cipher can be selected.',)'"],
        )

    def test_report_invalid_ssl_options(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_SSL_OPTIONS: "invalid"},
            specific_env_values={env.PCSD_SSL_OPTIONS: 0, "has_errors": True},
            errors=["Ignoring unknown SSL option 'invalid'"],
        )

    def test_report_invalid_ssl_options_warning(self):
        env.settings.default_ssl_options = "invalid"
        self.assert_environ_produces_modified_pcsd_env(
            specific_env_values={env.PCSD_SSL_OPTIONS: 0},
            warnings=["Ignoring unknown SSL option 'invalid'"],
        )

    def test_empty_bind_addresses(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_BIND_ADDR: " "},
            specific_env_values={env.PCSD_BIND_ADDR: {""}},
        )

    def test_no_disable_gui_explicitly(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_DISABLE_GUI: "false"},
        )

    def test_debug_explicitly(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_DEBUG: "true"},
            specific_env_values={env.PCSD_DEBUG: True},
        )

    def test_no_debug_explicitly(self):
        self.assert_environ_produces_modified_pcsd_env(
            {env.PCSD_DEBUG: "false"}
        )

    def test_errors_on_missing_paths(self):
        self.path_exists.return_value = False
        pcsd_dir = partial(join_path, settings.pcsd_exec_location)
        self.assert_environ_produces_modified_pcsd_env(
            specific_env_values={"has_errors": True},
            errors=[
                "Directory with web UI assets"
                f" '{pcsd_dir(env.PCSD_STATIC_FILES_DIR_NAME)}'"
                " does not exist",
            ],
        )

    def test_no_errors_on_missing_paths_disabled_gui(self):
        self.path_exists.return_value = False
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_DISABLE_GUI: "true"},
            specific_env_values={
                env.PCSD_DISABLE_GUI: True,
                "has_errors": False,
            },
            errors=[],
        )

    def test_invalid_worker_count(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_WORKER_COUNT: "0"},
            specific_env_values={
                env.PCSD_WORKER_COUNT: settings.pcsd_worker_count,
                "has_errors": True,
            },
            errors=[
                f"Value '0' for '{env.PCSD_WORKER_COUNT}' is not a positive integer"
            ],
        )

    def test_invalid_worker_reset_limit(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_WORKER_RESET_LIMIT: "a"},
            specific_env_values={
                env.PCSD_WORKER_RESET_LIMIT: settings.pcsd_worker_reset_limit,
                "has_errors": True,
            },
            errors=[
                f"Value 'a' for '{env.PCSD_WORKER_RESET_LIMIT}' is not a positive integer"
            ],
        )

    def test_invalid_max_worker_count(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_MAX_WORKER_COUNT: "0"},
            specific_env_values={
                env.PCSD_MAX_WORKER_COUNT: settings.pcsd_worker_count
                + settings.pcsd_temporary_workers,
                "has_errors": True,
            },
            errors=[
                f"Value '0' for '{env.PCSD_MAX_WORKER_COUNT}' is not a positive integer"
            ],
        )

    def test_max_worker_count_dependent_on_worker_count(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_WORKER_COUNT: "1"},
            specific_env_values={
                env.PCSD_WORKER_COUNT: 1,
                env.PCSD_MAX_WORKER_COUNT: 1 + settings.pcsd_temporary_workers,
            },
        )

    def test_invalid_deadlock_threshold_timeout(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_DEADLOCK_THRESHOLD_TIMEOUT: "-1"},
            specific_env_values={
                env.PCSD_DEADLOCK_THRESHOLD_TIMEOUT: settings.pcsd_deadlock_threshold_timeout,
                "has_errors": True,
            },
            errors=[
                f"Value '-1' for '{env.PCSD_DEADLOCK_THRESHOLD_TIMEOUT}' is not a non-negative integer"
            ],
        )

    def test_invalid_check_interval(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_CHECK_INTERVAL_MS: "0"},
            specific_env_values={
                env.PCSD_CHECK_INTERVAL_MS: settings.async_api_scheduler_interval_ms,
                "has_errors": True,
            },
            errors=[
                f"Value '0' for '{env.PCSD_CHECK_INTERVAL_MS}' is not a positive integer"
            ],
        )

    def test_invalid_task_abandoned_timeout(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_TASK_ABANDONED_TIMEOUT: "0"},
            specific_env_values={
                env.PCSD_TASK_ABANDONED_TIMEOUT: settings.task_abandoned_timeout_seconds,
                "has_errors": True,
            },
            errors=[
                f"Value '0' for '{env.PCSD_TASK_ABANDONED_TIMEOUT}' is not a positive integer"
            ],
        )

    def test_invalid_task_unresponsive_interval(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_TASK_UNRESPONSIVE_TIMEOUT: "0"},
            specific_env_values={
                env.PCSD_TASK_UNRESPONSIVE_TIMEOUT: settings.task_unresponsive_timeout_seconds,
                "has_errors": True,
            },
            errors=[
                f"Value '0' for '{env.PCSD_TASK_UNRESPONSIVE_TIMEOUT}' is not a positive integer"
            ],
        )

    def test_invalid_task_deletion_timeout(self):
        self.assert_environ_produces_modified_pcsd_env(
            environ={env.PCSD_TASK_DELETION_TIMEOUT: "-1"},
            specific_env_values={
                env.PCSD_TASK_DELETION_TIMEOUT: settings.task_deletion_timeout_seconds,
                "has_errors": True,
            },
            errors=[
                f"Value '-1' for '{env.PCSD_TASK_DELETION_TIMEOUT}' is not a non-negative integer"
            ],
        )
